import unittest
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

import openmdao.api as om
import dymos as dm
import numpy as np


@use_tempdirs
class TestDocOscillator(unittest.TestCase):

    def test_matrix_param(self):

        from dymos.examples.oscillator.oscillator_ode import OscillatorVectorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options["optimizer"] = 'SLSQP'

        static_params = False

        t = dm.Radau(num_segments=2, order=3)
        phase = dm.Phase(ode_class=OscillatorVectorODE, transcription=t,
                         ode_init_kwargs={'static_params': static_params})

        phase.set_time_options(fix_initial=True, duration_bounds=(1, 2), duration_ref=1)
        phase.add_state("x", fix_initial=True, rate_source="x_dot")

        A_mat = np.array(
            [
                [0, 1],
                [-1, 0]
            ]
        )

        # argument "dynamic" doesn't seem to help
        phase.add_parameter("A", val=A_mat, targets=["A"], static_target=static_params)
        phase.add_objective("time", loc="final", scaler=1)

        traj = dm.Trajectory()
        traj.add_phase("phase0", phase)

        prob.model.add_subsystem("traj", traj)

        prob.driver.declare_coloring()
        prob.setup(force_alloc_complex=True)
        phase.set_state_val('x', vals=[[1, 0], [1, 0]])

        dm.run_problem(prob, run_driver=True, simulate=True, make_plots=True)
        t_f = prob.get_val('traj.phase0.timeseries.time')[-1]
        final_state = prob.get_val('traj.phase0.timeseries.x')[-1, :]
        assert_near_equal(final_state, np.array([np.cos(t_f), -np.sin(t_f)]).ravel(),
                          tolerance=1e-5)

    def test_matrix_static_param(self):

        from dymos.examples.oscillator.oscillator_ode import OscillatorVectorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options["optimizer"] = 'SLSQP'

        static_params = True

        t = dm.Radau(num_segments=2, order=3)
        phase = dm.Phase(ode_class=OscillatorVectorODE, transcription=t,
                         ode_init_kwargs={'static_params': static_params})

        phase.set_time_options(fix_initial=True, duration_bounds=(1, 2), duration_ref=1)
        phase.add_state("x", fix_initial=True, rate_source="x_dot")

        A_mat = np.array(
            [
                [0, 1],
                [-1, 0]
            ]
        )

        # argument "dynamic" doesn't seem to help
        phase.add_parameter("A", val=A_mat, targets=["A"], static_target=static_params)
        phase.add_objective("time", loc="final", scaler=1)

        traj = dm.Trajectory()
        traj.add_phase("phase0", phase)

        prob.model.add_subsystem("traj", traj)

        prob.driver.declare_coloring()
        prob.setup(force_alloc_complex=True)
        phase.set_state_val('x', vals=[[1, 0], [1, 0]])

        dm.run_problem(prob, run_driver=True, simulate=True, make_plots=True)
        t_f = prob.get_val('traj.phase0.timeseries.time')[-1]
        final_state = prob.get_val('traj.phase0.timeseries.x')[-1, :]
        assert_near_equal(final_state, np.array([np.cos(t_f), -np.sin(t_f)]).ravel(),
                          tolerance=1e-5)



from .continuity_comp import RadauPSContinuityComp, GaussLobattoContinuityComp
from .control_comp import ControlInterpComp
from .parameter_comp import ParameterComp
from .time_comp import TimeComp
from .timeseries_output_comp import TimeseriesOutputComp



import unittest

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal

import numpy as np
import openmdao.api as om
import dymos as dm

g = 1.61544  # lunar gravity, m/s**2


class LaunchVehicle2DEOM(om.ExplicitComponent):
    """
    Simple 2D Cartesian Equations of Motion for a launch vehicle subject to thrust and drag.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('vx',
                       val=np.zeros(nn),
                       desc='x velocity',
                       units='m/s')

        self.add_input('vy',
                       val=np.zeros(nn),
                       desc='y velocity',
                       units='m/s')

        self.add_input('m',
                       val=np.zeros(nn),
                       desc='mass',
                       units='kg')

        self.add_input('theta',
                       val=np.zeros(nn),
                       desc='pitch angle',
                       units='rad')

        self.add_input('thrust',
                       val=2100000 * np.ones(nn),
                       desc='thrust',
                       units='N')

        self.add_input('Isp',
                       val=265.2 * np.ones(nn),
                       desc='specific impulse',
                       units='s')

        # Outputs
        self.add_output('xdot',
                        val=np.zeros(nn),
                        desc='velocity component in x',
                        units='m/s')

        self.add_output('ydot',
                        val=np.zeros(nn),
                        desc='velocity component in y',
                        units='m/s')

        self.add_output('vxdot',
                        val=np.zeros(nn),
                        desc='x acceleration magnitude',
                        units='m/s**2')

        self.add_output('vydot',
                        val=np.zeros(nn),
                        desc='y acceleration magnitude',
                        units='m/s**2')

        self.add_output('mdot',
                        val=np.zeros(nn),
                        desc='mass rate of change',
                        units='kg/s')

        # Setup partials
        ar = np.arange(self.options['num_nodes'])

        self.declare_partials(of='xdot', wrt='vx', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='ydot', wrt='vy', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='vxdot', wrt='vx', rows=ar, cols=ar)
        self.declare_partials(of='vxdot', wrt='m', rows=ar, cols=ar)
        self.declare_partials(of='vxdot', wrt='theta', rows=ar, cols=ar)
        self.declare_partials(of='vxdot', wrt='thrust', rows=ar, cols=ar)

        self.declare_partials(of='vydot', wrt='m', rows=ar, cols=ar)
        self.declare_partials(of='vydot', wrt='theta', rows=ar, cols=ar)
        self.declare_partials(of='vydot', wrt='vy', rows=ar, cols=ar)
        self.declare_partials(of='vydot', wrt='thrust', rows=ar, cols=ar)

        self.declare_partials(of='mdot', wrt='thrust', rows=ar, cols=ar)
        self.declare_partials(of='mdot', wrt='Isp', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        vx = inputs['vx']
        vy = inputs['vy']
        m = inputs['m']
        F_T = inputs['thrust']
        Isp = inputs['Isp']

        outputs['xdot'] = vx
        outputs['ydot'] = vy
        outputs['vxdot'] = F_T * cos_theta / m
        outputs['vydot'] = F_T * sin_theta / m - g
        outputs['mdot'] = -F_T / (g * Isp)

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        m = inputs['m']
        F_T = inputs['thrust']
        Isp = inputs['Isp']

        # jacobian['vxdot', 'vx'] = -CDA * rho * vx / m
        jacobian['vxdot', 'm'] = -(F_T * cos_theta) / m ** 2
        jacobian['vxdot', 'theta'] = -(F_T / m) * sin_theta
        jacobian['vxdot', 'thrust'] = cos_theta / m

        # jacobian['vydot', 'vy'] = -CDA * rho * vy / m
        jacobian['vydot', 'm'] = -(F_T * sin_theta) / m ** 2
        jacobian['vydot', 'theta'] = (F_T / m) * cos_theta
        jacobian['vydot', 'thrust'] = sin_theta / m

        jacobian['mdot', 'thrust'] = -1.0 / (g * Isp)
        jacobian['mdot', 'Isp'] = F_T / (g * Isp ** 2)


class LinearTangentGuidanceComp(om.ExplicitComponent):
    """ Compute pitch angle from static controls governing linear expression for
        pitch angle tangent as function of time.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('a_ctrl',
                       val=np.zeros(nn),
                       desc='linear tangent slope',
                       units='1/s')

        self.add_input('b_ctrl',
                       val=np.zeros(nn),
                       desc='tangent of theta at t=0',
                       units=None)

        self.add_input('time_phase',
                       val=np.zeros(nn),
                       desc='time',
                       units='s')

        self.add_output('theta',
                        val=np.zeros(nn),
                        desc='pitch angle',
                        units='rad')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='theta', wrt='a_ctrl', rows=arange, cols=arange, val=1.0)
        self.declare_partials(of='theta', wrt='b_ctrl', rows=arange, cols=arange, val=1.0)
        self.declare_partials(of='theta', wrt='time_phase', rows=arange, cols=arange, val=1.0)

    def compute(self, inputs, outputs):
        a = inputs['a_ctrl']
        b = inputs['b_ctrl']
        t = inputs['time_phase']
        outputs['theta'] = np.arctan(a * t + b)

    def compute_partials(self, inputs, jacobian):
        a = inputs['a_ctrl']
        b = inputs['b_ctrl']
        t = inputs['time_phase']

        x = a * t + b
        denom = x ** 2 + 1.0

        jacobian['theta', 'a_ctrl'] = t / denom
        jacobian['theta', 'b_ctrl'] = 1.0 / denom
        jacobian['theta', 'time_phase'] = a / denom


class LaunchVehicleLinearTangentODE(om.Group):
    """
    The LaunchVehicleLinearTangentODE for this case consists of a guidance component and
    the EOM.  Guidance is simply an OpenMDAO ExecComp which computes the arctangent of the
    tan_theta variable.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

    def setup(self):
        nn = self.options['num_nodes']
        self.add_subsystem('guidance', LinearTangentGuidanceComp(num_nodes=nn))
        self.add_subsystem('eom', LaunchVehicle2DEOM(num_nodes=nn))
        self.connect('guidance.theta', 'eom.theta')


@use_tempdirs
class TestSSTOLinearTangentTrajParams(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_invalid_traj_param_target_phase(self):
        #
        # Setup and solve the optimal control problem
        #
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)

        phase = dm.Phase(ode_class=LaunchVehicleLinearTangentODE,
                         transcription=dm.GaussLobatto(num_segments=10, order=5, compressed=True))

        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(10, 1000),
                               targets=['guidance.time_phase'])

        phase.add_state('x', fix_initial=True, lower=0, rate_source='eom.xdot', units='m')
        phase.add_state('y', fix_initial=True, lower=0, rate_source='eom.ydot', units='m')
        phase.add_state('vx', fix_initial=True, lower=0, rate_source='eom.vxdot', targets=['eom.vx'], units='m/s')
        phase.add_state('vy', fix_initial=True, rate_source='eom.vydot', targets=['eom.vy'], units='m/s')
        phase.add_state('m', fix_initial=True, rate_source='eom.mdot', targets=['eom.m'], units='kg')

        phase.add_boundary_constraint('y', loc='final', equals=1.85E5, linear=True)
        phase.add_boundary_constraint('vx', loc='final', equals=1627.0)
        phase.add_boundary_constraint('vy', loc='final', equals=0)

        traj.add_parameter('a_ctrl', units='1/s', opt=True, targets={'foo': ['guidance.a_ctrl']})
        traj.add_parameter('b_ctrl', units=None, opt=True, targets={'phase0': ['guidance.b_ctrl']})
        phase.add_parameter('thrust', units='N', opt=False, val=3.0 * 50000.0 * 1.61544, targets=['eom.thrust'])
        phase.add_parameter('Isp', units='s', opt=False, val=1.0E6, targets=['eom.Isp'])

        phase.add_objective('time', index=-1, scaler=0.01)

        p.model.linear_solver = om.DirectSolver()

        phase.add_timeseries_output('guidance.theta', units='deg')

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = "The following phases were specified as\n" \
                   "targets for trajectory parameter a_ctrl\n" \
                   "but are not valid phase names: {'foo'}"

        self.assertEqual(expected, str(e.exception))

    @require_pyoptsparse(optimizer='SLSQP')
    def test_opt_traj_param_ode_targets(self):
        #
        # Setup and solve the optimal control problem
        #
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)

        phase = dm.Phase(ode_class=LaunchVehicleLinearTangentODE,
                         transcription=dm.GaussLobatto(num_segments=10, order=5, compressed=True))

        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(10, 1000),
                               targets=['guidance.time_phase'])

        phase.add_state('x', fix_initial=True, lower=0, rate_source='eom.xdot', units='m')
        phase.add_state('y', fix_initial=True, lower=0, rate_source='eom.ydot', units='m')
        phase.add_state('vx', fix_initial=True, lower=0, rate_source='eom.vxdot', targets=['eom.vx'], units='m/s')
        phase.add_state('vy', fix_initial=True, rate_source='eom.vydot', targets=['eom.vy'], units='m/s')
        phase.add_state('m', fix_initial=True, rate_source='eom.mdot', targets=['eom.m'], units='kg')

        phase.add_boundary_constraint('y', loc='final', equals=1.85E5, linear=True)
        phase.add_boundary_constraint('vx', loc='final', equals=1627.0)
        phase.add_boundary_constraint('vy', loc='final', equals=0)

        traj.add_parameter('a_ctrl', units='1/s', opt=True, targets={'phase0': ['guidance.a_ctrl']})
        traj.add_parameter('b_ctrl', units=None, opt=True, targets={'phase0': ['guidance.b_ctrl']})
        phase.add_parameter('thrust', units='N', opt=False, val=3.0 * 50000.0 * 1.61544, targets=['eom.thrust'])
        phase.add_parameter('Isp', units='s', opt=False, val=1.0E6, targets=['eom.Isp'])

        phase.add_objective('time', index=-1, scaler=0.01)

        p.model.linear_solver = om.DirectSolver()

        phase.add_timeseries_output('guidance.theta', units='deg')

        p.setup(check=True)

        phase.set_time_val(initial=0.0, duration=500.0)
        phase.set_state_val('x', [0, 350000.0])
        phase.set_state_val('y', [0, 185000.0])
        phase.set_state_val('vx', [0, 1627.0])
        phase.set_state_val('vy', [1.0E-6, 0.0])
        phase.set_state_val('m', 50000)
        phase.set_parameter_val('a_ctrl', -0.01)
        phase.set_parameter_val('b_ctrl', 3.0)

        dm.run_problem(p, simulate=True)

        a = p.get_val('traj.parameter_vals:a_ctrl')
        b = p.get_val('traj.parameter_vals:b_ctrl')

        assert_near_equal(a, [[-0.00828051]], tolerance=1.0E-5)
        assert_near_equal(b, [[2.74739928]], tolerance=1.0E-5)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from numpy.testing import assert_almost_equal

try:
    import jax
    import jax.numpy as jnp
except ImportError as e:
    jax = None
    jnp = None

import openmdao.api as om
import dymos as dm

from openmdao.utils.testing_utils import use_tempdirs
from dymos.utils.misc import om_version


class GuidedBrachistochroneODE(om.JaxExplicitComponent):
    """
    The brachistochrone EOM assuming a linear rate of change in the wire angle wrt time.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # static parameters
        self.add_input('v', desc='speed of the bead on the wire', shape=(nn,), units='m/s')
        self.add_input('g', desc='gravitational acceleration', shape=(1,), units='m/s**2')
        self.add_input('time_phase', desc='phase elapsed time', shape=(nn,), units='s')
        self.add_input('theta_rate', desc='guidance parameter', shape=(1,), units='rad/s')

        self.add_output('xdot', desc='velocity component in x', shape=(nn,), units='m/s',
                        tags=['dymos.state_rate_source:x', 'dymos.state_units:m'])

        self.add_output('ydot', desc='velocity component in y', shape=(nn,), units='m/s',
                        tags=['dymos.state_rate_source:y', 'dymos.state_units:m'])

        self.add_output('vdot', desc='acceleration magnitude', shape=(nn,), units='m/s**2',
                        tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])

        self.add_output('theta', desc='wire angle', shape=(nn,), val=2, units='rad')

        # TODO: Temporary pending bug fix in JaxExplicitComp
        self.declare_coloring(method='jax')

    # because our compute primal output depends on static variables, in this case
    # and self.options['num_nodes'], we must define a get_self_statics method. This method must
    # return a tuple of all static variables. Their order in the tuple doesn't matter.  If your
    # component happens to have discrete inputs, do NOT return them here. Discrete inputs are passed
    # into the compute_primal function individually, after the continuous variables.
    def get_self_statics(self):
        # return value must be hashable
        return self.options['num_nodes'],

    def compute_primal(self, v, g, time_phase, theta_rate):
        theta = time_phase * theta_rate

        sin_theta = jnp.sin(theta)
        cos_theta = jnp.cos(theta)

        vdot = g * cos_theta
        xdot = v * sin_theta
        ydot = -v * cos_theta

        return xdot, ydot, vdot, theta


@unittest.skipIf(jax is None, 'Test Requires Jax')
@unittest.skipIf(om_version()[0] <= (3, 37, 0), 'Requires OpenMDAO version later than 3.37.0')
@use_tempdirs
class TestBrachistochroneBoundaryBalance(unittest.TestCase):

    def make_problem(self, tx):
        p = om.Problem(model=om.Group())

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=GuidedBrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_parameter('g', units='m/s**2')
        phase.add_parameter('theta_rate', units='deg/s')

        phase.add_calc_expr('v2 = v * v')

        phase.add_timeseries_output('theta', units='deg')
        phase.add_timeseries_output('v2', units='m**2/s**2')

        phase.add_boundary_balance(param='t_duration', name='x', tgt_val=10.0, loc='final', lower=0.1, upper=5.0)
        phase.add_boundary_balance(param='theta_rate', name='y', tgt_val=5.0, loc='final', lower=0.1, upper=100.0, res_ref=1.0)

        phase.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, maxiter=100, iprint=0, atol=1.0E-6, rtol=1.0E-6)
        phase.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        phase.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=1.8016)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_parameter_val('theta_rate',  20.0)
        phase.set_parameter_val('g', 9.80665)

        p.final_setup()

        return p

    def run_asserts(self, p):

        t_initial = p.get_val('traj0.phase0.timeseries.time')[0]
        tf = p.get_val('traj0.phase0.timeseries.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.x')[0]
        xf = p.get_val('traj0.phase0.timeseries.x')[-1]

        y0 = p.get_val('traj0.phase0.timeseries.y')[0]
        yf = p.get_val('traj0.phase0.timeseries.y')[-1]

        v0 = p.get_val('traj0.phase0.timeseries.v')[0]
        vf = p.get_val('traj0.phase0.timeseries.v')[-1]

        g = p.get_val('traj0.phase0.parameter_vals:g')[0]

        thetaf = p.get_val('traj0.phase0.timeseries.theta')[-1]

        v2_f = p.get_val('traj0.phase0.timeseries.v2')[-1]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(tf, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(v2_f, vf**2, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

        outputs = [op[1]['prom_name'] for op in p.model.list_outputs(out_stream=None, prom_name=True)]
        self.assertNotIn('traj0.phase0.timeseries.theta_rate', outputs)

    def test_picard(self):
        for grid_type in 'cgl', 'lgl':
            with self.subTest(f'{grid_type=}'):
                tx = dm.PicardShooting(num_segments=1, nodes_per_seg=21, grid_type='cgl')
                p = self.make_problem(tx=tx)

                p.run_model()

                self.run_asserts(p)

    def test_birkhoff(self):
        for grid_type in 'cgl', 'lgl':
            with self.subTest(f'{grid_type=}'):
                tx = dm.Birkhoff(num_nodes=21, grid_type=grid_type, solve_segments='forward')
                p = self.make_problem(tx=tx)

                p.run_model()

                self.run_asserts(p)

    def test_radau(self):
        tx = dm.Radau(num_segments=10, order=3, solve_segments='forward')
        p = self.make_problem(tx=tx)

        p.run_model()

        self.run_asserts(p)

    def test_gausslobatto(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments='forward')
        p = self.make_problem(tx=tx)

        p.run_model()

        self.run_asserts(p)

    def test_explicit_shooting(self):
        tx = dm.ExplicitShooting(num_segments=10, order=3)
        p = self.make_problem(tx=tx)

        p.run_model()

        self.run_asserts(p)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND

from dymos.models.atmosphere.atmos_1976 import USatm1976Data


english_to_metric_rho = om.unit_conversion('slug/ft**3', 'kg/m**3')[0]
english_to_metric_h = om.unit_conversion('ft', 'm')[0]
rho_interp = InterpND(points=np.array(USatm1976Data.alt*english_to_metric_h),
                      values=np.array(USatm1976Data.rho*english_to_metric_rho),
                      method='slinear').interpolate


class CannonballODE(om.ExplicitComponent):
    """
    Cannonball ODE assuming flat earth and accounting for air resistance
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # static parameters
        self.add_input('m', units='kg')
        self.add_input('S', units='m**2')
        # 0.5 good assumption for a sphere
        self.add_input('CD', 0.5)

        # time varying inputs
        self.add_input('h', units='m', shape=nn)
        self.add_input('v', units='m/s', shape=nn)
        self.add_input('gam', units='rad', shape=nn)

        # state rates
        self.add_output('v_dot', shape=nn, units='m/s**2')
        self.add_output('gam_dot', shape=nn, units='rad/s')
        self.add_output('h_dot', shape=nn, units='m/s')
        self.add_output('r_dot', shape=nn, units='m/s')
        self.add_output('ke', shape=nn, units='J')

        # Ask OpenMDAO to compute the partial derivatives using complex-step
        # with a partial coloring algorithm for improved performance, and use
        # coloring to determine the sparsity pattern.
        self.declare_coloring(wrt='*', method='cs')

    def compute(self, inputs, outputs):

        gam = inputs['gam']
        v = inputs['v']
        h = inputs['h']
        m = inputs['m']
        S = inputs['S']
        CD = inputs['CD']

        GRAVITY = 9.80665  # m/s**2

        rho = rho_interp(h)

        q = 0.5*rho*inputs['v']**2
        qS = q * S
        D = qS * CD
        cgam = np.cos(gam)
        sgam = np.sin(gam)
        outputs['v_dot'] = - D/m-GRAVITY*sgam
        outputs['gam_dot'] = -(GRAVITY/v)*cgam
        outputs['h_dot'] = v*sgam
        outputs['r_dot'] = v*cgam
        outputs['ke'] = 0.5*m*v**2



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
import dymos as dm

from dymos.utils.testing_utils import SimpleODE, SimpleVectorizedODE
from dymos.transcriptions.explicit_shooting.ode_evaluation_group import ODEEvaluationGroup


class TestODEEvaluationGroup(unittest.TestCase):

    def test_eval(self):
        ode_class = SimpleODE
        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 's**2'
        state_options['x']['rate_source'] = 'x_dot'
        state_options['x']['targets'] = ['x']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}

        p = om.Problem()

        igd = dm.GaussLobattoGrid(num_segments=1, nodes_per_seg=3, compressed=False)

        p.model.add_subsystem('ode_eval', ODEEvaluationGroup(ode_class,
                                                             input_grid_data=igd,
                                                             time_options=time_options,
                                                             state_options=state_options,
                                                             parameter_options=param_options,
                                                             control_options=control_options,
                                                             ode_init_kwargs=None))
        p.setup(force_alloc_complex=True)

        p.model.ode_eval.set_segment_index(0)
        p.set_val('ode_eval.states:x', [1.25])
        p.set_val('ode_eval.time', [2.2])
        p.set_val('ode_eval.parameters:p', [1.0])

        p.run_model()

        x = p.get_val('ode_eval.states:x')
        t = p.get_val('ode_eval.time')
        xdot_check = x - t**2 + 1

        assert_near_equal(p.get_val('ode_eval.state_rate_collector.state_rates:x_rate'), xdot_check)

        cpd = p.check_partials(compact_print=True, method='cs', out_stream=None)
        assert_check_partials(cpd)

    def test_eval_vectorized(self):
        ode_class = SimpleVectorizedODE
        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'z': dm.phase.options.StateOptionsDictionary()}

        state_options['z']['shape'] = (2,)
        state_options['z']['units'] = 's**2'
        state_options['z']['rate_source'] = 'z_dot'
        state_options['z']['targets'] = ['z']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}

        p = om.Problem()

        igd = dm.GaussLobattoGrid(num_segments=1, nodes_per_seg=3, compressed=False)

        p.model.add_subsystem('ode_eval', ODEEvaluationGroup(ode_class,
                                                             input_grid_data=igd,
                                                             time_options=time_options,
                                                             state_options=state_options,
                                                             parameter_options=param_options,
                                                             control_options=control_options,
                                                             ode_init_kwargs=None))
        p.setup(check=False, force_alloc_complex=True)

        p.model.ode_eval.set_segment_index(0)
        p.set_val('ode_eval.states:z', [1.25, 0.0])
        p.set_val('ode_eval.time', [2.2])
        p.set_val('ode_eval.parameters:p', [1.0])

        p.run_model()

        z = p.get_val('ode_eval.states:z')
        p_ = p.get_val('ode_eval.parameters:p')
        t = p.get_val('ode_eval.time')

        z_rate = p.get_val('ode_eval.state_rate_collector.state_rates:z_rate')

        assert_near_equal(z_rate[0, 0], z[:, 0] - t**2 + p_)
        assert_near_equal(z_rate[0, 1], 10 * t)

        cpd = p.check_partials(compact_print=True, method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import time
import unittest

import numpy as np

import openmdao.api as om

import dymos
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.misc import GroupWrapperConfig
from dymos.utils.testing_utils import PhaseStub, SimpleODE
from dymos.transcriptions.common.time_comp import TimeComp
from dymos.transcriptions.picard_shooting.multiple_shooting_iter_group import MultipleShootingIterGroup
from dymos.phase.options import StateOptionsDictionary, TimeOptionsDictionary
from dymos.transcriptions.grid_data import GaussLobattoGrid, ChebyshevGaussLobattoGrid


TimeComp = GroupWrapperConfig(TimeComp, [])
MultipleShootingIterGroupWrapped = GroupWrapperConfig(MultipleShootingIterGroup, [PhaseStub()])


class LorenzAttractorODE(om.JaxExplicitComponent):
    """
    The ODE for the Lorenz attractor.
    """

    def initialize(self):
        """
        All Dymos ODE systems are required to have an option "num_nodes",
        which is the number of points at which the ODE is simultaneously evaluated.

        This will be set by the Phase during setup once the transcription details are known.
        """
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        """
        In setup, we add inputs and outputs.

        The first dimension is assumed to pertain to the index of the node.

        An input that's a scalar at each node should have a shape of
        (num_nodes, 1) or just (num_nodes,).

        For vectors or matrices, it's just the shape of the matrix at each
        node prepended with num_nodes.

        We provide units for the scalars, but OpenMDAO doesn't do unit conversion on an index-by-index basis,
        so we just assume that no unit conversion should be done for the S matrix and K vector.
        """
        nn = self.options['num_nodes']

        # ODE inputs
        self.add_input('x', shape=(nn,))
        self.add_input('y', shape=(nn,))
        self.add_input('z', shape=(nn,))
        self.add_input('s', shape=(1,))
        self.add_input('r', shape=(1,))
        self.add_input('b', shape=(1,))

        # State rates
        self.add_output('x_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:x'])
        self.add_output('y_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:y'])
        self.add_output('z_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:z'])

    # because our compute primal output depends on static variables, in this case
    # and self.options['num_nodes'], we must define a get_self_statics method. This method must
    # return a tuple of all static variables. Their order in the tuple doesn't matter.  If your
    # component happens to have discrete inputs, do NOT return them here. Discrete inputs are passed
    # into the compute_primal function individually, after the continuous variables.
    def get_self_statics(self):
        # return value must be hashable
        return self.options['num_nodes'],

    def setup_partials(self):
        nn = self.options['num_nodes']
        ar = np.arange(nn, dtype=int)
        self.declare_partials('x_dot', 'x', rows=ar, cols=ar)
        self.declare_partials('x_dot', 'y', rows=ar, cols=ar)
        self.declare_partials('x_dot', 's')

        self.declare_partials('y_dot', 'x', rows=ar, cols=ar)
        self.declare_partials('y_dot', 'y', rows=ar, cols=ar)
        self.declare_partials('y_dot', 'z', rows=ar, cols=ar)
        self.declare_partials('y_dot', 'r')

        self.declare_partials('z_dot', 'x', rows=ar, cols=ar)
        self.declare_partials('z_dot', 'y', rows=ar, cols=ar)
        self.declare_partials('z_dot', 'z', rows=ar, cols=ar)
        self.declare_partials('z_dot', 'b')

    def compute_primal(self, x, y, z, s=10, r=28, b=2.667):
        """
        Parameters
        ----------
        x : array-like
            The x state.
        y : array-like
            The y state.
        z : array-like
            The z state.
        s : float
            The Lorenz s parameter.
        r : float
            The Lorenz r parameter.
        b : float
            The Lorenz b parameter.

        Returns
        -------
        x_dot : array-like
            The time-derivative of the x state.
        y_dot : array-like
            The time-derivative of the y state.
        z_dot : array-like
            The time-derivative of the z state.
        """
        x_dot = s * (y - x)
        y_dot = r * x - y - x*z
        z_dot = x * y - b * z
        return x_dot, y_dot, z_dot


@use_tempdirs
class TestMultipleShootingIterGroup(unittest.TestCase):

    def test_multiple_shooting_iter_group(self):
        for direction in ['forward', 'backward']:
            for grid_type in [GaussLobattoGrid, ChebyshevGaussLobattoGrid]:
                for num_seg, nodes_per_seg in [(1, 21), (3, 11)]:
                    grid_data = grid_type(nodes_per_seg=nodes_per_seg, num_segments=num_seg)
                    for nl_solver in [om.NonlinearBlockGS(use_aitken=True, maxiter=100, iprint=0),
                                      om.NewtonSolver(solve_subsystems=True, maxiter=100, iprint=0)]:
                        with self.subTest(msg=f'{direction=} grid={grid_data}'):
                            with dymos.options.temporary(include_check_partials=True):

                                state_options = {'x': StateOptionsDictionary()}

                                state_options['x']['shape'] = (1,)
                                state_options['x']['units'] = 's**2'
                                state_options['x']['targets'] = ['x']
                                state_options['x']['initial_bounds'] = (None, None)
                                state_options['x']['final_bounds'] = (None, None)
                                state_options['x']['solve_segments'] = direction
                                state_options['x']['rate_source'] = 'x_dot'

                                time_options = TimeOptionsDictionary()
                                ode_class = SimpleODE

                                p = om.Problem()

                                class PhaseStub(om.Group):

                                    def setup(self):
                                        self.state_options = state_options
                                        self.time_options = time_options

                                    def classify_var(self, var):
                                        return 'ode'

                                    def configure(self):
                                        self._get_subsystem('ms').configure_io(self)
                                        self._get_subsystem('time').configure_io()

                                p.model = PhaseStub()

                                p.model.add_subsystem('time', TimeComp(num_nodes=grid_data.num_nodes,
                                                                       node_ptau=grid_data.node_ptau,
                                                                       node_dptau_dstau=grid_data.node_dptau_dstau,
                                                                       units='s'))

                                p.model.add_subsystem('ms', MultipleShootingIterGroup(state_options=state_options,
                                                                                      time_units=time_options['units'],
                                                                                      grid_data=grid_data,
                                                                                      ode_class=ode_class,
                                                                                      ms_nonlinear_solver=nl_solver),
                                                      promotes=['ode_all*'])
                                p.model.connect('time.t', 'ode_all.t')
                                p.model.connect('time.dt_dstau', 'ms.picard_update_comp.dt_dstau')

                                ms = p.model._get_subsystem('ms')

                                ms.nonlinear_solver = nl_solver
                                ms.linear_solver = om.DirectSolver()

                                p.setup(force_alloc_complex=True)

                                p.set_val('time.t_initial', 0.0)
                                p.set_val('time.t_duration', 2.0)

                                def solution(t):
                                    return t**2 + 2 * t + 1 - 0.5 * np.exp(t)

                                def dsolution_dt(t):
                                    return 2 * t + 2 - 0.5 * np.exp(t)

                                if direction == 'forward':
                                    p.set_val('ms.seg_initial_states:x', 0.5)
                                    p.set_val('ms.initial_states:x', 0.5)
                                else:
                                    p.set_val('ms.seg_final_states:x', solution(2.0))
                                    p.set_val('ms.final_states:x', solution(2.0))

                                p.set_val('ode_all.p', 1.0)

                                p.final_setup()

                                t_start = time.perf_counter()
                                p.run_model()
                                t_end = time.perf_counter()

                                # print(f"Elapsed time: {t_end-t_start:.4f} seconds")

                                t = p.get_val('time.t')
                                x = p.get_val('ms.states:x')
                                x_dot = p.get_val('ode_all.x_dot')

                                assert_near_equal(solution(t), x.ravel(), tolerance=1.0E-9)
                                assert_near_equal(dsolution_dt(t), x_dot.ravel(), tolerance=1.0E-9)
                                assert_near_equal(solution(0), p.get_val('ms.initial_states:x').ravel(), tolerance=1.0E-9)
                                assert_near_equal(solution(t[-1]), p.get_val('ms.final_states:x').ravel(), tolerance=1.0E-9)

                                cpd = p.check_partials(method='fd', compact_print=False, out_stream=None)
                                assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.transcriptions.grid_data import GridData
from dymos.transcriptions.pseudospectral.components import CollocationComp
from dymos.transcriptions.pseudospectral.components.state_independents import StateIndependentsComp
from dymos.utils.testing_utils import assert_check_partials

# Modify class so we can run it standalone.
from dymos.utils.misc import CompWrapperConfig
CollocationComp = CompWrapperConfig(CollocationComp)
StateIndependentsComp = CompWrapperConfig(StateIndependentsComp)


@use_tempdirs
class TestCollocationCompSolOpt(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def _make_state_idx_map(self, state_name, options, grid_data, state_idx_map):
        """
        Provides error checking for solve_segments and establishes necessary data structures.

        Parameters
        ----------
        state_name : str
            The name of the state being configured.
        options : StateOptionsDictionary
            The StateOptionsDictionary for the state being configured.
        grid_data : GridData
            The GridData object which holds the grid information for this test.
        state_idx_map : dict
            The state_idx_map to be populated by this method.  This map is equivalent to the
            one generated by PseudospectralBase._configure_solve_segments.
        """
        if state_idx_map is None:
            state_idx_map = {}
        state_idx_map[state_name] = {'solver': None, 'indep': None}

        state_input_idxs = grid_data.subset_node_indices['state_input']
        num_state_input_nodes = grid_data.subset_num_nodes['state_input']
        compressed = options.get('compressed', False)

        # Sanity-checks for solve segments
        # If solve_segments is used at all, we cannot fix the state at both ends of the phase.
        if options['solve_segments']:
            # Forward propagation
            if options['solve_segments'] in {True, 'forward'}:
                if compressed:
                    state_idx_map[state_name]['solver'] = np.arange(1, num_state_input_nodes, dtype=int)
                    state_idx_map[state_name]['indep'] = np.zeros((1,), dtype=int)
                else:
                    left_idxs = grid_data.subset_node_indices['segment_ends'][0::2]
                    state_idx_map[state_name]['solver'] = [i for i in range(num_state_input_nodes)
                                                           if state_input_idxs[i] not in left_idxs]
                    state_idx_map[state_name]['indep'] = [i for i in range(num_state_input_nodes)
                                                          if state_input_idxs[i] in left_idxs]

            # Backward propagation
            elif options['solve_segments'] in {'backward'}:
                if compressed:
                    # The optimizer controls the last state input node, all others are solver-controlled
                    state_idx_map[state_name]['indep'] = np.array([num_state_input_nodes - 1], dtype=int)
                    state_idx_map[state_name]['solver'] = np.arange(num_state_input_nodes - 1, dtype=int)
                else:
                    # The optimizer controls the last state input node in each segment, all others are solver_controlled
                    right_idxs = grid_data.subset_node_indices['segment_ends'][1::2]
                    state_idx_map[state_name]['indep'] = [i for i in range(num_state_input_nodes)
                                                          if state_input_idxs[i] in right_idxs]
                    state_idx_map[state_name]['solver'] = [i for i in range(num_state_input_nodes)
                                                           if state_input_idxs[i] not in right_idxs]
        else:
            # No solver used to solve these nodes.  All state input nodes are the indep nodes.
            state_idx_map[state_name]['solver'] = []
            state_idx_map[state_name]['indep'] = np.arange(len(state_input_idxs), dtype=int)

    def make_prob(self, transcription, n_segs, order, compressed):

        p = om.Problem(model=om.Group())

        gd = GridData(num_segments=n_segs, segment_ends=np.arange(n_segs+1),
                      transcription=transcription, transcription_order=order, compressed=compressed)

        state_options = {'x': {'units': 'm', 'shape': (1, ), 'fix_initial': True,
                               'fix_final': False, 'solve_segments': False,
                               'input_initial': False, 'val': 1.0, 'lower': None,
                               'upper': None},
                         'v': {'units': 'm/s', 'shape': (3, 2), 'fix_initial': False,
                               'fix_final': True, 'solve_segments': True,
                               'input_initial': False, 'val': np.ones((3, 2)),
                               'lower': None, 'upper': None}}

        indep_comp = om.IndepVarComp()
        p.model.add_subsystem('indep', indep_comp, promotes_outputs=['*'])

        indep_comp.add_output(
            'dt_dstau',
            val=np.zeros((gd.subset_num_nodes['col']))
        )

        indep_comp.add_output(
            'f_approx:x',
            val=np.zeros((gd.subset_num_nodes['col'], 1)), units='m')
        indep_comp.add_output(
            'f_computed:x',
            val=np.zeros((gd.subset_num_nodes['col'], 1)), units='m')

        indep_comp.add_output(
            'f_approx:v',
            val=np.zeros((gd.subset_num_nodes['col'], 3, 2)), units='m/s')
        indep_comp.add_output(
            'f_computed:v',
            val=np.zeros((gd.subset_num_nodes['col'], 3, 2)), units='m/s')

        p.model.add_subsystem('defect_comp',
                              subsys=CollocationComp(grid_data=gd, state_options=state_options))

        indep = StateIndependentsComp(grid_data=gd, state_options=state_options)
        p.model.add_subsystem('state_indep', indep, promotes_outputs=['*'])

        p.model.connect('f_approx:x', 'defect_comp.f_approx:x')
        p.model.connect('f_approx:v', 'defect_comp.f_approx:v')
        p.model.connect('f_computed:x', 'defect_comp.f_computed:x')
        p.model.connect('f_computed:v', 'defect_comp.f_computed:v')
        p.model.connect('dt_dstau', 'defect_comp.dt_dstau')
        p.model.connect('defect_comp.defects:v', 'state_indep.defects:v')

        self.state_idx_map = {}
        for state_name, options in state_options.items():
            self._make_state_idx_map(state_name, options, gd, self.state_idx_map)
        p.model.state_indep.configure_io(self.state_idx_map)

        p.setup(force_alloc_complex=True)

        p['dt_dstau'] = np.random.random(gd.subset_num_nodes['col'])

        p['f_approx:x'] = np.random.random((gd.subset_num_nodes['col'], 1))
        p['f_approx:v'] = np.random.random((gd.subset_num_nodes['col'], 3, 2))

        p['f_computed:x'] = np.random.random((gd.subset_num_nodes['col'], 1))
        p['f_computed:v'] = np.random.random((gd.subset_num_nodes['col'], 3, 2))

        p.run_model()
        p.model.run_apply_nonlinear()

        return p

    def test_results(self):
        p = self.make_prob('gauss-lobatto', n_segs=4, order=3, compressed=False)

        dt_dstau = p['dt_dstau']

        assert_almost_equal(p['defect_comp.defects:x'],
                            dt_dstau[:, np.newaxis] * (p['f_approx:x']-p['f_computed:x']))

        # solver_nodes = p.model.state_indep.solver_node_idx[:-1]  # fix_final
        solver_nodes = self.state_idx_map['v']['solver']
        assert_almost_equal(p.model._residuals['state_indep.states:v'][solver_nodes],
                            dt_dstau[:, np.newaxis, np.newaxis] *
                            (p['f_approx:v']-p['f_computed:v']))

        p = self.make_prob('gauss-lobatto', n_segs=4, order=3, compressed=True)

        dt_dstau = p['dt_dstau']

        assert_almost_equal(p['defect_comp.defects:x'],
                            dt_dstau[:, np.newaxis] * (p['f_approx:x']-p['f_computed:x']))

        solver_nodes = self.state_idx_map['v']['solver']
        assert_almost_equal(p.model._residuals['state_indep.states:v'][solver_nodes],
                            dt_dstau[:, np.newaxis, np.newaxis] *
                            (p['f_approx:v']-p['f_computed:v']))

    def test_partials(self):
        np.set_printoptions(linewidth=1024, edgeitems=1e1000)

        p = self.make_prob('radau-ps', n_segs=2, order=5, compressed=False)
        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        del cpd['state_indep']
        assert_check_partials(cpd)

        p = self.make_prob('radau-ps', n_segs=2, order=5, compressed=True)
        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        del cpd['state_indep']
        assert_check_partials(cpd)

        p = self.make_prob('gauss-lobatto', n_segs=3, order=5, compressed=False)
        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        del cpd['state_indep']
        assert_check_partials(cpd)

        p = self.make_prob('gauss-lobatto', n_segs=4, order=3, compressed=True)
        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        del cpd['state_indep']
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



"""Test Dymos Linkage Reports GUI using Playwright."""
import asyncio
from packaging.version import Version
from playwright.async_api import async_playwright
from aiounittest import async_test
import unittest

from openmdao.utils.gui_testing_utils import _GuiTestCase
import sys

import openmdao.api as om
import openmdao
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp
from dymos.visualization.linkage.report import create_linkage_report

if 'win32' in sys.platform:
    # Windows specific event-loop policy & cmd
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

my_loop = asyncio.new_event_loop()
asyncio.set_event_loop(my_loop)

""" A set of toolbar tests that runs on each model. """
toolbar_script = [
    {
        "desc": "Uncollapse All button",
        "id": "expand-all",
    },
    {
        "desc": "Collapse Outputs in View Only button",
        "id": "collapse-element-2",
    },
    {
        "desc": "Uncollapse In View Only button",
        "id": "expand-element",
    },
    {
        "desc": "Show Legend (off) button",
        "id": "legend-button",
    },
    {
        "desc": "Show Legend (on) button",
        "id": "legend-button",
    },
    {
        "desc": "Show Path (on) button",
        "id": "info-button",
    },
    {
        "desc": "Show Path (off) button",
        "id": "info-button",
    },
    {
        "desc": "Clear Arrows and Connection button",
        "id": "hide-connections",
    },
    {
        "desc": "Help (on) button",
        "id": "question-button",
    },
    {
        "desc": "Help (off) button",
        "id": "question-button",
    },
    {
        "desc": "Collapse All Outputs button",
        "id": "collapse-all",
    }
]

""" A dictionary of tests script with an array for each model."""
gui_test_script = [
    {"test": "toolbar"},
    {
        "desc": "Hover on matrix cell and check arrow count",
        "test": "hoverArrow",
        "selector": "g#n2elements rect#cellShape_node_41.vMid",
        "arrowCount": 2
    },
    {
        "desc": "Left-click on model tree element to zoom",
        "test": "click",
        "selector": "g#tree g.phase rect#climb",
        "button": "left"
    },
    # TODO: This particular test is broken due to timeout, not sure why yet.
    # {
    #     "desc": "Hover on matrix cell and check arrow count",
    #     "test": "hoverArrow",
    #     "selector": "g#n2elements rect#cellShape_node_128.vMid",
    #     "arrowCount": 1
    # },
    {"test": "root"},
    {
        "desc": "Right-click on model tree element to collapse",
        "test": "click",
        "selector": "g#tree rect#br_to_v1",
        "button": "right"
    },
    {
        "desc": "Hover over collapsed matrix cell and check arrow count",
        "test": "hoverArrow",
        "selector": "g#n2elements rect#cellShape_node_18.gMid",
        "arrowCount": 19
    },
    {
        "desc": "Right-click on model tree element to uncollapse",
        "test": "click",
        "selector": "g#tree rect#br_to_v1",
        "button": "right"
    },
    {"test": "root"},
    {
        "desc": "Check the number of cells in the matrix grid",
        "test": "count",
        "selector": "g#n2elements > g.n2cell",
        "count": 196
    },
    {
        "desc": "Perform a search on states:v",
        "test": "search",
        "searchString": "states:v",
        "diagElementCount": 25
    },
    {"test": "root"},
    {
        "desc": "Check that home button works after search",
        "test": "count",
        "selector": "g#n2elements > g.n2cell",
        "count": 196
    },
    {
        "desc": "Expand toolbar connections menu",
        "test": "hover",
        "selector": ".group-3 > div.expandable:first-child"
    },
    {
        "desc": "Press toolbar show all connections button",
        "test": "click",
        "selector": "#show-all-connections",
        "button": "left"
    },
    {
        "desc": "Check number of arrows",
        "test": "count",
        "selector": "g#n2arrows > g",
        "count": 154
    },
    {
        "desc": "Expand toolbar connections menu",
        "test": "hover",
        "selector": ".group-3 > div.expandable:first-child"
    },
    {
        "desc": "Press toolbar hide all connections button",
        "test": "click",
        "selector": "#hide-connections-2",
        "button": "left"
    },
    {
        "desc": "Check number of arrows",
        "test": "count",
        "selector": "g#n2arrows > g",
        "count": 0
    },
    {"test": "root"},
    {
        "desc": "Alt-right-click the climb.initial component",
        "test": "click",
        "selector": "rect#climb_initial",
        "button": "right",
        "modifiers": ["Alt"]
    },
    {
        "desc": "Check that variable selection dialog appears",
        "test": "count",
        "selector": "#childSelect-climb_initial",
        "count": 1
    },
    {
        "desc": "Select a variable to hide",
        "test": "click",
        "selector": "input#climb_initial_time-visible-check",
        "button": "left"
    },
    {
        "desc": "Click the Apply button",
        "test": "click",
        "selector": ".button-container button:last-child",
        "button": "left"
    },
    {
        "desc": "Check that the time variable is no longer displayed",
        "test": "count",
        "selector": "rect#climb_initial_time",
        "count": 0
    },
    {
        "desc": "Alt-right-click the climb.initial component",
        "test": "click",
        "selector": "rect#climb_initial",
        "button": "right",
        "modifiers": ["Alt"]
    },
    {
        "desc": "Click the Clear Search button",
        "test": "click",
        "selector": ".search-clear",
        "button": "left"
    },
    {
        "desc": "Perform a search in the variable selection dialog",
        "test": "var_select_search",
        "searchString": "gam",
        "foundVariableCount": 1
    },
    {
        "desc": "Click the Apply button",
        "test": "click",
        "selector": ".button-container button:last-child",
        "button": "left"
    },
    {
        "desc": "Check the number of cells in the matrix grid",
        "test": "count",
        "selector": "g#n2elements > g.n2cell",
        "count": 192
    },

]

resize_dirs = {
    'top': [0, -1],
    'top-right': [1, -1],
    'right': [1, 0],
    'bottom-right': [1, 1],
    'bottom': [0, 1],
    'bottom-left': [-1, 1],
    'left': [-1, 0],
    'top-left': [-1, -1]
}

current_test = 1


@unittest.skipUnless(Version(openmdao.__version__) > Version("3.19"), "reports API is too old")
@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class dymos_linkage_gui_test_case(_GuiTestCase):
    def setUp(self):
        p = om.Problem()

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['derivative_test'] = 'first-order'

        p.driver.declare_coloring()

        # First Phase: Brake release to V1 - both engines operable
        br_to_v1 = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
        br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
        br_to_v1.add_state('v', fix_initial=True, lower=0.0001, ref=100.0, defect_ref=100.0)
        br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
        br_to_v1.add_timeseries_output('*')

        # Second Phase: Rejected takeoff at V1 - no engines operable
        rto = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                       ode_init_kwargs={'mode': 'runway'})
        rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rto.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
        rto.add_timeseries_output('*')

        # Third Phase: V1 to Vr - single engine operable
        v1_to_vr = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        v1_to_vr.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
        v1_to_vr.add_timeseries_output('*')

        # Fourth Phase: Rotate - single engine operable
        rotate = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                          ode_init_kwargs={'mode': 'runway'})
        rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
        rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rotate.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        rotate.add_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10,
                           ref=10, val=[0, 10], control_type='polynomial')
        rotate.add_timeseries_output('*')

        # Fifth Phase: Climb to target speed and altitude at end of runway.
        climb = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=5),
                         ode_init_kwargs={'mode': 'climb'})
        climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
        climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        climb.add_state('h', fix_initial=True, lower=0.0, ref=1.0, defect_ref=1.0)
        climb.add_state('v', fix_initial=False, lower=0.0001, ref=100.0, defect_ref=100.0)
        climb.add_state('gam', fix_initial=True, lower=0.0, ref=0.05, defect_ref=0.05)
        climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
        climb.add_timeseries_output('*')

        # Instantiate the trajectory and add phases
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)
        traj.add_phase('br_to_v1', br_to_v1)
        traj.add_phase('rto', rto)
        traj.add_phase('v1_to_vr', v1_to_vr)
        traj.add_phase('rotate', rotate)
        traj.add_phase('climb', climb)

        # Add parameters common to multiple phases to the trajectory
        traj.add_parameter('m', val=174200., opt=False, units='lbm',
                           desc='aircraft mass',
                           targets={'br_to_v1': ['m'], 'v1_to_vr': ['m'], 'rto': ['m'],
                                    'rotate': ['m'], 'climb': ['m']})

        traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                           desc='nominal aircraft thrust',
                           targets={'br_to_v1': ['T']})

        traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                           desc='thrust under a single engine',
                           targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

        traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                           desc='thrust when engines are shut down for rejected takeoff',
                           targets={'rto': ['T']})

        traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                           desc='nominal runway friction coeffcient',
                           targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'], 'rotate': ['mu_r']})

        traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                           desc='runway friction coefficient under braking',
                           targets={'rto': ['mu_r']})

        traj.add_parameter('h_runway', val=0., opt=False, units='ft', static_target=False,
                           desc='runway altitude',
                           targets={'br_to_v1': ['h'], 'v1_to_vr': ['h'], 'rto': ['h'],
                                    'rotate': ['h']})

        traj.add_parameter('rho', val=1.225, opt=False, units='kg/m**3', static_target=True,
                           desc='atmospheric density',
                           targets={'br_to_v1': ['rho'], 'v1_to_vr': ['rho'], 'rto': ['rho'],
                                    'rotate': ['rho']})

        traj.add_parameter('S', val=124.7, opt=False, units='m**2', static_target=True,
                           desc='aerodynamic reference area',
                           targets={'br_to_v1': ['S'], 'v1_to_vr': ['S'], 'rto': ['S'],
                                    'rotate': ['S'], 'climb': ['S']})

        traj.add_parameter('CD0', val=0.03, opt=False, units=None, static_target=True,
                           desc='zero-lift drag coefficient',
                           targets={f'{phase}': ['CD0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('AR', val=9.45, opt=False, units=None, static_target=True,
                           desc='wing aspect ratio',
                           targets={f'{phase}': ['AR'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                     'rto', 'rotate', 'climb']})

        traj.add_parameter('e', val=801, opt=False, units=None, static_target=True,
                           desc='Oswald span efficiency factor',
                           targets={f'{phase}': ['e'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                    'rto', 'rotate', 'climb']})

        traj.add_parameter('span', val=35.7, opt=False, units='m', static_target=True,
                           desc='wingspan',
                           targets={f'{phase}': ['span'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                       'rto', 'rotate', 'climb']})

        traj.add_parameter('h_w', val=1.0, opt=False, units='m', static_target=True,
                           desc='height of wing above CG',
                           targets={f'{phase}': ['h_w'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL0', val=0.5, opt=False, units=None, static_target=True,
                           desc='zero-alpha lift coefficient',
                           targets={f'{phase}': ['CL0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL_max', val=2.0, opt=False, units=None, static_target=True,
                           desc='maximum lift coefficient for linear fit',
                           targets={f'{phase}': ['CL_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                         'rto', 'rotate', 'climb']})

        # Standard "end of first phase to beginning of second phase" linkages
        traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
        traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

        # Less common "final value of r must be the match at ends of two phases".
        traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                    phase_b='climb', var_b='r', loc_b='final',
                                    ref=1000,
                                    connected=False)

        # Define the constraints and objective for the optimal control problem
        rto.add_boundary_constraint('v', loc='final', upper=0.001, ref=100, linear=True)

        rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

        climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
        climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
        climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
        climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=1.2)

        rto.add_objective('r', loc='final', ref=1000.0)

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p.set_val('traj.br_to_v1.t_initial', 0)
        p.set_val('traj.br_to_v1.t_duration', 35)
        p.set_val('traj.br_to_v1.states:r', br_to_v1.interp('r', [0, 2500.0]))
        p.set_val('traj.br_to_v1.states:v', br_to_v1.interp('v', [0.0001, 100.0]))
        p.set_val('traj.br_to_v1.parameters:alpha', 0, units='deg')

        p.set_val('traj.v1_to_vr.t_initial', 35)
        p.set_val('traj.v1_to_vr.t_duration', 35)
        p.set_val('traj.v1_to_vr.states:r', v1_to_vr.interp('r', [2500, 300.0]))
        p.set_val('traj.v1_to_vr.states:v', v1_to_vr.interp('v', [100, 110.0]))
        p.set_val('traj.v1_to_vr.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rto.t_initial', 35)
        p.set_val('traj.rto.t_duration', 1)
        p.set_val('traj.rto.states:r', rto.interp('r', [2500, 5000.0]))
        p.set_val('traj.rto.states:v', rto.interp('v', [110, 0.0001]))
        p.set_val('traj.rto.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rotate.t_initial', 35)
        p.set_val('traj.rotate.t_duration', 5)
        p.set_val('traj.rotate.states:r', rotate.interp('r', [1750, 1800.0]))
        p.set_val('traj.rotate.states:v', rotate.interp('v', [80, 85.0]))
        p.set_val('traj.rotate.controls:alpha', 0.0, units='deg')

        p.set_val('traj.climb.t_initial', 30)
        p.set_val('traj.climb.t_duration', 20)
        p.set_val('traj.climb.states:r', climb.interp('r', [5000, 5500.0]), units='ft')
        p.set_val('traj.climb.states:v', climb.interp('v', [160, 170.0]), units='kn')
        p.set_val('traj.climb.states:h', climb.interp('h', [0, 35.0]), units='ft')
        p.set_val('traj.climb.states:gam', climb.interp('gam', [0, 5.0]), units='deg')
        p.set_val('traj.climb.controls:alpha', 5.0, units='deg')

        create_linkage_report(traj)

    def log_test(self, msg):
        global current_test

        """ Print a description and index for the test about to run. """
        print("  Test {:04}".format(current_test) + ": " + msg)
        current_test += 1

    async def load_test_page(self):
        """ Load the specified HTML file from the local filesystem. """
        url = f'file://{self.tempdir}/linkage_report.html'

        # Without wait_until: 'networkidle', processing will begin before
        # the page is fully rendered
        await self.page.goto(url, wait_until='networkidle')

    async def generic_toolbar_tests(self):
        """ Click most of the toolbar buttons to see if an error occurs """
        for test in toolbar_script:
            with self.subTest(test['desc']):
                self.log_test("[Toolbar] " + test['desc'])

                btnHandle = await self.get_handle('#' + test['id'])
                await btnHandle.click(button='left', timeout=3333, force=True)

        await self.page.reload(wait_until='networkidle')

    async def assert_element_count(self, selector, expected_found):
        """
        Count the number of elements located by the selector and make
        sure it exactly matches the supplied value. Try several times
        because sometimes transition animations throw things off.
        """
        max_tries = 3  # Max number of times to attempt to find a selector
        max_time = 2000  # The timeout in ms for each search

        if (expected_found > 0):
            num_tries = 0
            found = False
            while (not found and num_tries < max_tries):
                nth_selector = f':nth-match({selector}, {expected_found})'
                try:
                    await self.page.wait_for_selector(nth_selector, state='attached',
                                                      timeout=max_time)
                    found = True
                except Exception:
                    num_tries += 1

            num_tries = 0
            found = False
            while (not found and num_tries < max_tries):
                nth_selector = f':nth-match({selector}, {expected_found + 1})'
                try:
                    await self.page.wait_for_selector(nth_selector, state='detached',
                                                      timeout=max_time)
                    found = True
                except Exception:
                    num_tries += 1

        else:
            num_tries = 0
            found = False
            while (not found and num_tries < max_tries):
                nth_selector = f':nth-match({selector}, 1)'
                try:
                    await self.page.wait_for_selector(nth_selector, state='detached',
                                                      timeout=max_time)
                    found = True
                except Exception:
                    num_tries += 1

        hndl_list = await self.page.query_selector_all(selector)
        if (len(hndl_list) > expected_found):
            global current_test
            await self.page.screenshot(path=f'shot_{current_test}.png')

        self.assertEqual(len(hndl_list), expected_found,
                         'Found ' + str(len(hndl_list)) +
                         ' elements, expected ' + str(expected_found))

    async def assert_arrow_count(self, expected_arrows):
        """
        Count the number of path elements in the n2arrows < div > and make
        sure it matches the specified value.
        """
        await self.assert_element_count('g#n2arrows > g', expected_arrows)

    async def hover(self, options, log_test=True):
        """
        Hover over the specified element.
        """
        if log_test:
            self.log_test(options['desc'] if 'desc' in options else
                          "Hover over '" + options['selector'] + "'")

        hndl = await self.get_handle(options['selector'])

        await hndl.hover(force=False)

    async def hover_and_check_arrow_count(self, options):
        """
        Hover over a matrix cell, make sure the number of expected arrows
        are there, then move off and make sure the arrows go away.
        """
        await self.hover(options)

        # Make sure there are enough arrows
        await self.assert_arrow_count(options['arrowCount'])
        await self.page.mouse.move(0, 0)  # Get the mouse off the element
        await self.assert_arrow_count(0)  # Make sure no arrows left

    async def click(self, options):
        """
        Perform a click of the type specified by options.button on the
        element specified by options.selector.
        """
        self.log_test(options['desc'] if 'desc' in options else
                      options['button'] + "-click on '" +
                      options['selector'] + "'")

        hndl = await self.get_handle(options['selector'])

        mod_keys = [] if 'modifiers' not in options else options['modifiers']
        await hndl.click(button=options['button'], modifiers=mod_keys)

    async def drag(self, options):
        """
        Hover over the element, perform a mousedown event, move the mouse to the
        specified location, and perform a mouseup. Check to make sure the element
        moved in at least one direction.
        """
        self.log_test(options['desc'] if 'desc' in options else
                      "Dragging '" + options['selector'] + "' to " + options['x'] + "," + options['y'])

        hndl = await self.get_handle(options['selector'])

        pre_drag_bbox = await hndl.bounding_box()

        await hndl.hover(force=True)
        await self.page.mouse.down()
        await self.page.mouse.move(options['x'], options['y'])
        await self.page.mouse.up()

        post_drag_bbox = await hndl.bounding_box()

        moved = ((pre_drag_bbox['x'] != post_drag_bbox['x']) or
                 (pre_drag_bbox['y'] != post_drag_bbox['y']))

        self.assertIsNot(moved, False,
                         "The '" + options['selector'] + "' element did not move.")

    async def resize_window(self, options):
        """
        Drag an edge/corner of a WindowResizable obj and check that the size changed
        or didn't change as expected.
        """
        self.log_test(options['desc'] if 'desc' in options else
                      "Resizing '" + options['selector'] + "' window.")

        win_hndl = await self.get_handle(options['selector'])
        pre_resize_bbox = await win_hndl.bounding_box()

        edge_hndl = await self.get_handle(options['selector'] + ' div.rsz-' + options['side'])
        edge_bbox = await edge_hndl.bounding_box()

        new_x = edge_bbox['x'] + resize_dirs[options['side']][0] * options['distance']
        new_y = edge_bbox['y'] + resize_dirs[options['side']][1] * options['distance']

        await edge_hndl.hover()
        await self.page.mouse.down()
        await self.page.mouse.move(new_x, new_y)
        await self.page.mouse.up()

        post_resize_bbox = await win_hndl.bounding_box()
        dw = post_resize_bbox['width'] - pre_resize_bbox['width']
        dh = post_resize_bbox['height'] - pre_resize_bbox['height']

        resized = ((dw != 0) or (dh != 0))
        if options['expectChange']:
            self.assertIsNot(resized, False,
                             "The '" + options['selector'] + "' element was NOT resized and should have been.")
        else:
            self.assertIsNot(resized, True,
                             "The '" + options['selector'] + "' element was resized and should NOT have been.")

    async def return_to_root(self):
        """
        Left-click the home button and wait for the transition to complete.
        """

        self.log_test("Return to root")
        hndl = await self.get_handle("#reset-graph")
        await hndl.click()

    async def search_and_check_result(self, options):
        """
        Enter a string in the search textbox and check that the expected
        number of elements are shown in the N2 matrix.
        """
        searchString = options['searchString']
        self.log_test(options['desc'] if 'desc' in options else
                      "Searching for '" + options['searchString'] +
                      "' and checking for " +
                      str(options['diagElementCount']) + " diagram elements after.")

        await self.page.click("#searchbar-container")

        searchbar = await self.page.wait_for_selector('#awesompleteId', state='visible')
        await searchbar.type(searchString + "\n", delay=50)

        await self.assert_element_count("g.n2cell", options['diagElementCount'])

    async def var_select_search_and_check_result(self, options):
        """
        Enter a string in the variable selection search textbox and check the result.
        """
        searchString = options['searchString']
        self.log_test(options['desc'] if 'desc' in options else
                      "Searching for '" + options['searchString'] +
                      "' and checking for " +
                      str(options['foundVariableCount']) + " table rows after.")

        searchbar = await self.page.wait_for_selector('.search-container input', state='visible')
        await searchbar.type(searchString + "\n", delay=50)

        await self.assert_element_count("td.varname", options['foundVariableCount'])

    async def run_model_script(self, script):
        """
        Iterate through the supplied script array and perform each
        action/test.
        """

        print("Running tests from model script...")

        for script_item in script:
            if 'test' not in script_item:
                continue

            test_type = script_item['test']
            if test_type == 'hoverArrow':
                await self.hover_and_check_arrow_count(script_item)
            elif test_type == 'hover':
                await self.hover(script_item)
            elif test_type == 'click':
                await self.click(script_item)
            elif test_type == 'drag':
                await self.drag(script_item)
            elif test_type == 'resize':
                await self.resize_window(script_item)
            elif test_type == 'root':
                await self.return_to_root()
            elif test_type == 'search':
                await self.search_and_check_result(script_item)
            elif test_type == 'var_select_search':
                await self.var_select_search_and_check_result(script_item)
            elif test_type == 'toolbar':
                await self.generic_toolbar_tests()
            elif test_type == 'count':
                self.log_test(script_item['desc'] if 'desc' in script_item
                              else "Checking for " + str(script_item['count']) +
                                   "' instances of '" + script_item['selector'] + "'")
                await self.assert_element_count(script_item['selector'],
                                                script_item['count'])

    async def run_gui_tests(self, playwright):
        """ Execute all of the tests in an async event loop. """
        await self.setup_browser(playwright)
        await self.load_test_page()
        await self.run_model_script(gui_test_script)

        await self.browser.close()

        if self.console_error:
            msg = "Console log contains errors."
            print(msg)
            self.fail(msg)

        if self.page_error:
            msg = "There were errors on the page."
            print(msg)
            self.fail(msg)

    @async_test(loop=my_loop)
    async def test_gui(self):
        print("\n" + '-' * 78 + "\n" + '-' * 78 + "\n" + '-' * 78)

        self.current_test_desc = ''
        self.current_model = 'bfl'

        async with async_playwright() as playwright:
            await self.run_gui_tests(playwright)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.testing_utils import assert_check_partials

import dymos as dm
from dymos.transcriptions.pseudospectral.components import GaussLobattoInterleaveComp
from dymos.transcriptions.grid_data import GridData


class TestGaussLobattoInterleaveComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

        self.grid_data = gd = GridData(num_segments=3, segment_ends=np.array([0., 2., 4., 10.0]),
                                       transcription='gauss-lobatto', transcription_order=[3, 3, 3])

        num_disc_nodes = gd.subset_num_nodes['state_disc']
        num_col_nodes = gd.subset_num_nodes['col']

        self.p = om.Problem(model=om.Group())

        state_options = {'u': {'units': 'm', 'shape': (1,)},
                         'v': {'units': 'm', 'shape': (3, 2)}}

        ode_outputs = {'vehicle_cg': {'units': 'm', 'shape': (3,)}}

        indep_comp = om.IndepVarComp()
        self.p.model.add_subsystem('indep', indep_comp, promotes=['*'])

        indep_comp.add_output('state_disc:u',
                              val=np.zeros((num_disc_nodes, 1)), units='m')

        indep_comp.add_output('state_disc:v',
                              val=np.zeros((num_disc_nodes, 3, 2)), units='m')

        indep_comp.add_output('state_col:u',
                              val=np.zeros((num_col_nodes, 1)), units='m')

        indep_comp.add_output('state_col:v',
                              val=np.zeros((num_col_nodes, 3, 2)), units='m')

        indep_comp.add_output('ode_disc:cg',
                              val=np.zeros((num_disc_nodes, 3)), units='m')

        indep_comp.add_output('ode_col:cg',
                              val=np.zeros((num_col_nodes, 3)), units='m')

        glic = self.p.model.add_subsystem('interleave_comp',
                                          subsys=GaussLobattoInterleaveComp(grid_data=gd))

        glic.add_var('u', **state_options['u'], disc_src='state_disc:u', col_src='state_col:u')
        glic.add_var('v', **state_options['v'], disc_src='state_disc:v', col_src='state_col:v')
        glic.add_var('vehicle_cg', **ode_outputs['vehicle_cg'], disc_src='ode_disc:cg', col_src='ode_col:cg')

        self.p.model.connect('state_disc:u', 'interleave_comp.disc_values:u')
        self.p.model.connect('state_disc:v', 'interleave_comp.disc_values:v')
        self.p.model.connect('state_col:u', 'interleave_comp.col_values:u')
        self.p.model.connect('state_col:v', 'interleave_comp.col_values:v')

        self.p.model.connect('ode_disc:cg', 'interleave_comp.disc_values:vehicle_cg')
        self.p.model.connect('ode_col:cg', 'interleave_comp.col_values:vehicle_cg')

        self.p.setup(force_alloc_complex=True)

        self.p['state_disc:u'] = np.random.random((num_disc_nodes, 1))
        self.p['state_disc:v'] = np.random.random((num_disc_nodes, 3, 2))
        self.p['state_col:u'] = np.random.random((num_col_nodes, 1))
        self.p['state_col:v'] = np.random.random((num_col_nodes, 3, 2))

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        u_disc = self.p.get_val('state_disc:u')
        v_disc = self.p.get_val('state_disc:v')
        u_col = self.p.get_val('state_col:u')
        v_col = self.p.get_val('state_col:v')

        u_all = self.p.get_val('interleave_comp.all_values:u')
        v_all = self.p.get_val('interleave_comp.all_values:v')

        assert_near_equal(u_all[self.grid_data.subset_node_indices['state_disc'], ...],
                          u_disc)

        assert_near_equal(v_all[self.grid_data.subset_node_indices['state_disc'], ...],
                          v_disc)

        assert_near_equal(u_all[self.grid_data.subset_node_indices['col'], ...],
                          u_col)

        assert_near_equal(v_all[self.grid_data.subset_node_indices['col'], ...],
                          v_col)

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=True, method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import importlib
import os
import shutil
import unittest
import pathlib

try:
    import matplotlib
except ImportError:
    matplotlib = None

from openmdao.utils.testing_utils import require_pyoptsparse
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs, set_env_vars_context

import dymos as dm
from dymos.examples.finite_burn_orbit_raise.finite_burn_orbit_raise_problem import two_burn_orbit_raise_problem
from dymos.utils.testing_utils import _get_reports_dir

_, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)
bokeh_available = importlib.util.find_spec('bokeh') is not None


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestExampleTwoBurnOrbitRaise(unittest.TestCase):

    def setUp(self):
        # We need to remove the TESTFLO_RUNNING environment variable for reports to be generated.
        # The reports code checks to see if TESTFLO_RUNNING is set and will not do anything if set
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)

    def tearDown(self):
        if os.path.isdir('plots'):
            shutil.rmtree('plots')

        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    @unittest.skipIf(not bokeh_available, 'bokeh unavailable')
    def test_bokeh_plots(self):
        with set_env_vars_context(OPENMDAO_REPORTS='1'):
            with dm.options.temporary(plots='bokeh'):
                p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                                 compressed=False, optimizer='SLSQP', show_output=False,
                                                 run_driver=False)

                html_file = pathlib.Path(_get_reports_dir(p)) / 'traj_results_report.html'
                self.assertTrue(html_file.exists(), msg=f'{html_file} does not exist!')

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_mpl_plots(self):
        with dm.options.temporary(plots='matplotlib'):
            p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                             compressed=False, optimizer='SLSQP', show_output=False,
                                             run_driver=False)

            expected_files = ('deltav.png', 'r.png', 'accel.png',
                              'u1.png', 'vr.png', 'pos_x.png',
                              'vt.png', 'pos_y.png', 'theta.png')

            for file in expected_files:
                plotfile = pathlib.Path(p.get_reports_dir(p)).joinpath('plots') / file
                self.assertTrue(plotfile.exists(), msg=f'{plotfile} does not exist!')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



"""Define the StateIndependents class."""

import numpy as np

import openmdao.api as om

from ....transcriptions.grid_data import GridData
from ...._options import options as dymos_options


class StateIndependentsComp(om.ImplicitComponent):
    """
    Class definition for the StateIndependentsComp.

    A simple component that replaces the state IndepVarComps whenever the solver needs to solve for
    the state or whenever the initial state is connected to an external source.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

    def configure_io(self, state_idx_map=None):
        """
        I/O creation is delayed until configure so we can determine shape and units.

        Parameters
        ----------
        state_idx_map : dict
            The state_idx_map generated by the transcription during its configuration.  This dict
            is keyed by state, which in turn holds a dict of the solve indices in the input state
            vector (keyed as 'solved') and those indices managed by the optimizer (keyed as 'indep').
            In some unittest, this map is not provided at the time of setup and is instead called
            later.
        """
        if state_idx_map is None:
            return
        self.state_idx_map = state_idx_map
        state_options = self.options['state_options']
        grid_data = self.options['grid_data']
        num_col_nodes = grid_data.subset_num_nodes['col']

        num_state_input_nodes = grid_data.subset_num_nodes['state_input']

        self.var_names = {}
        for state_name in state_options:
            self.var_names[state_name] = {
                'defect': f'defects:{state_name}',
            }

        for state_name, options in state_options.items():

            shape = options['shape']
            units = options['units']
            solved = options['solve_segments']
            default_val = options['val']
            if np.isscalar(default_val) or np.asarray(default_val).shape == 1:
                default_val = float(default_val)
            elif np.asarray(default_val).shape == shape:
                default_val = np.repeat(default_val[np.newaxis, ...], num_state_input_nodes, axis=0)
            var_names = self.var_names[state_name]

            if solved and (options['lower'] or options['upper']):
                om.issue_warning(f'State {state_name} has bounds but they are not enforced when '
                                 f'using `solve_segments.` Apply a path constraint to {state_name} '
                                 f'to enforce bounds.', om.UnusedOptionWarning)

            # only need the implicit variable if this state is solved.
            # Note: we don't add scaling and bounds here. This may be revisited.
            self.add_output(name=f'states:{state_name}',
                            shape=(num_state_input_nodes,) + shape,
                            val=default_val,
                            units=units)

            # Input for continuity, which can come from an external source.
            if options['input_initial']:
                input_name = f'initial_states:{state_name}'
                self.add_input(name=input_name, shape=(1, ) + shape, units=units)

            # compute an output constraint value since the optimizer needs it
            if solved:
                self.add_input(
                    name=var_names['defect'],
                    shape=(num_col_nodes, ) + shape,
                    desc=f'Constraint value for interior defects of state {state_name}',
                    units=units)

        # Setup partials
        for state_name, options in state_options.items():
            shape = options['shape']
            size = np.prod(shape)
            solved = options['solve_segments']
            state_var_name = f'states:{state_name}'

            if solved:  # only need this deriv if its solved
                solve_idx = np.array(state_idx_map[state_name]['solver'])
                indep_idx = np.array(state_idx_map[state_name]['indep'])

                num_indep_nodes = indep_idx.shape[0]
                num_solve_nodes = solve_idx.shape[0]

                base_idx = np.tile(np.arange(size), num_indep_nodes).reshape(num_indep_nodes, size)
                row = (indep_idx[:, np.newaxis]*size + base_idx).flatten()

                # anything that looks like an indep
                self.declare_partials(of=state_var_name, wrt=state_var_name,
                                      rows=row, cols=row, val=-1.0)

                if options['input_initial']:
                    wrt = f'initial_states:{state_name}'
                    row_col = np.arange(np.prod(shape))
                    self.declare_partials(of=state_var_name, wrt=wrt, rows=row_col, cols=row_col,
                                          val=1.0)

                col = np.arange(num_solve_nodes * size)
                base_idx = np.tile(np.arange(size), num_solve_nodes).reshape(num_solve_nodes, size)
                row = (solve_idx[:, np.newaxis]*size + base_idx).flatten()

                var_names = self.var_names[state_name]
                self.declare_partials(of=state_var_name,
                                      wrt=var_names['defect'],
                                      rows=row, cols=col, val=1.0)

            else:
                row_col = np.arange(num_state_input_nodes*np.prod(shape), dtype=int)
                self.declare_partials(of=state_var_name, wrt=state_var_name,
                                      rows=row_col, cols=row_col, val=-1.0)

                if options['input_initial']:
                    wrt = f'initial_states:{state_name}'
                    row_col = np.arange(np.prod(shape))
                    self.declare_partials(of=state_var_name, wrt=wrt, rows=row_col, cols=row_col,
                                          val=1.0)

    def apply_nonlinear(self, inputs, outputs, residuals):
        """
        Compute residuals given inputs and outputs.

        The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        residuals : Vector
            Unscaled, dimensional residuals written to via residuals[key].
        """
        state_options = self.options['state_options']

        for state_name, options in state_options.items():
            state_var_name = f'states:{state_name}'

            solve_idx = self.state_idx_map[state_name]['solver']
            indep_idx = self.state_idx_map[state_name]['indep']

            var_names = self.var_names[state_name]

            if options['solve_segments']:
                defect = inputs[var_names['defect']]
                residuals[state_var_name][solve_idx, ...] = defect

            # really is: <idep_val> - \outputs[state_name][indep_idx] but OpenMDAO
            # implementation details mean we just set it to 0
            # but derivatives are still based on (<idep_val> - \outputs[state_name][indep_idx]),
            # so you get -1 wrt state var
            # NOTE: check_partials will report wrong derivs for the indep vars,
            #       but don't believe it!
            residuals[state_var_name][indep_idx, ...] = 0.0

            if options['input_initial']:
                ic_state_name = f'initial_states:{state_name}'

                residuals[state_var_name][0, ...] = \
                    inputs[ic_state_name][0, ...] - \
                    outputs[state_var_name][0, ...]

    def solve_nonlinear(self, inputs, outputs):
        """
        Compute outputs given inputs.

        The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        """
        state_options = self.options['state_options']

        for state_name, options in state_options.items():
            if options['input_initial']:
                output_name = f'states:{state_name}'
                input_name = f'initial_states:{state_name}'
                outputs[output_name][0, ...] = inputs[input_name]



import unittest
import warnings

import numpy as np

import openmdao.api as om
import dymos as dm
from dymos import options as dymos_options

from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


class Simple2StateODE(om.ExplicitComponent):
    """
    A simple ODE from https://math.okstate.edu/people/yqwang/teaching/math4513_fall11/Notes/rungekutta.pdf
    """
    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('x', shape=(nn,), units='s**2')
        self.add_input('y', shape=(nn,), units='s')
        self.add_input('t', shape=(nn,), units='s')
        self.add_input('p', shape=(nn,), units='s**2')

        self.add_output('x_dot', shape=(nn,), units='s')
        self.add_output('y_dot', shape=(nn,), units=None)

        ar = np.arange(nn, dtype=int)
        self.declare_partials(of='x_dot', wrt='x', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='x_dot', wrt='t', rows=ar, cols=ar)
        self.declare_partials(of='x_dot', wrt='p', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='y_dot', wrt='y', rows=ar, cols=ar, val=-2.0)
        self.declare_partials(of='y_dot', wrt='t', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        x = inputs['x']
        y = inputs['y']
        t = inputs['t']
        p = inputs['p']
        outputs['x_dot'] = x - t**2 + p
        outputs['y_dot'] = -2 * y + t**3 * np.exp(-2*t)

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        partials['x_dot', 't'] = -2*t
        partials['y_dot', 't'] = -np.exp(-2*t) * t**2 * (2 * t - 3)


class Simple1StateODE(om.ExplicitComponent):
    """
    A simple ODE from https://math.okstate.edu/people/yqwang/teaching/math4513_fall11/Notes/rungekutta.pdf
    """
    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('y', shape=(nn,), units=None)
        self.add_input('t', shape=(nn,), units='s')

        self.add_output('y_dot', shape=(nn,), units='1/s')

        ar = np.arange(nn, dtype=int)

        self.declare_partials(of='y_dot', wrt='y', rows=ar, cols=ar, val=-2.0)
        self.declare_partials(of='y_dot', wrt='t', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        y = inputs['y']
        t = inputs['t']
        outputs['y_dot'] = -2 * y + t**3 * np.exp(-2*t)

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        partials['y_dot', 't'] = -np.exp(-2*t) * t**2 * (2 * t - 3)


@use_tempdirs
class TestExplicitShooting(unittest.TestCase):

    def test_1_state_run_model(self):

        dymos_options['include_check_partials'] = True

        prob = om.Problem()

        phase = dm.Phase(ode_class=Simple1StateODE,
                         transcription=dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=1,
                                                                                    nodes_per_seg=3,
                                                                                    compressed=True)))

        phase.set_time_options(targets=['t'], units='s')

        # automatically discover states
        phase.set_state_options('y', targets=['y'], rate_source='y_dot')

        prob.model.add_subsystem('phase0', phase)

        prob.setup(force_alloc_complex=False)

        prob.set_val('phase0.t_initial', 0.0)
        prob.set_val('phase0.t_duration', 1.0)
        prob.set_val('phase0.initial_states:y', 1.0)

        prob.run_model()

        with np.printoptions(linewidth=1024):
            cpd = prob.check_partials(compact_print=True, out_stream=None)

        assert_check_partials(cpd, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    def test_2_states_run_model(self):

        dymos_options['include_check_partials'] = True

        prob = om.Problem()

        phase = dm.Phase(ode_class=Simple2StateODE,
                         transcription=dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=2,
                                                                                    nodes_per_seg=3,
                                                                                    compressed=True)))

        phase.set_time_options(targets=['t'], units='s')

        # automatically discover states
        phase.set_state_options('x', targets=['x'], rate_source='x_dot')
        phase.set_state_options('y', targets=['y'], rate_source='y_dot')

        phase.add_parameter('p', targets=['p'])

        prob.model.add_subsystem('phase0', phase)

        prob.setup(force_alloc_complex=True)

        prob.set_val('phase0.t_initial', 0.0)
        prob.set_val('phase0.t_duration', 1.0)
        prob.set_val('phase0.initial_states:x', 0.5)
        prob.set_val('phase0.initial_states:y', 1.0)
        prob.set_val('phase0.parameters:p', 1)

        prob.run_model()

        t_f = prob.get_val('phase0.integrator.t_final')
        x_f = prob.get_val('phase0.integrator.states_out:x')
        y_f = prob.get_val('phase0.integrator.states_out:y')

        assert_near_equal(t_f, 1.0)
        assert_near_equal(x_f[-1, ...], 2.64085909, tolerance=1.0E-5)
        assert_near_equal(y_f[-1, ...], 0.1691691, tolerance=1.0E-5)

        with np.printoptions(linewidth=1024):
            cpd = prob.check_partials(compact_print=True, method='fd', out_stream=None)
            assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    def test_brachistochrone_explicit_shooting(self):

        dymos_options['include_check_partials'] = True

        input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=False)

        output_grids = {'same': input_grid,
                        'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7, compressed=True),
                        'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                        'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):

            with self.subTest(f'output_grid = {output_grid_type}'):
                prob = om.Problem()

                phase = dm.Phase(ode_class=BrachistochroneODE,
                                 transcription=dm.ExplicitShooting(grid=input_grid,
                                                                   output_grid=output_grids[output_grid_type]))

                traj = prob.model.add_subsystem('traj0', dm.Trajectory())
                traj.add_phase('phase0', phase)

                prob.driver = om.ScipyOptimizeDriver()

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9,
                                  ref=90., rate_continuity=True, rate2_continuity=False)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('traj0.phase0.t_initial', 0.0)
                prob.set_val('traj0.phase0.t_duration', 2)
                prob.set_val('traj0.phase0.initial_states:x', 0.0)
                prob.set_val('traj0.phase0.initial_states:y', 10.0)
                prob.set_val('traj0.phase0.initial_states:v', 0.1)
                prob.set_val('traj0.phase0.parameters:g', 9.80665, units='m/s**2')
                prob.set_val('traj0.phase0.controls:theta', phase.interp('theta', ys=[10, 50]), units='deg')

                # prob.run_model()

                dm.run_problem(prob)

                x = prob.get_val('traj0.phase0.timeseries.x')
                y = prob.get_val('traj0.phase0.timeseries.y')
                t = prob.get_val('traj0.phase0.timeseries.time')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)
                assert_near_equal(t[-1, ...], 1.8016, tolerance=1.0E-2)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=True, method='cs',
                                              excludes=['traj0.phases.phase0.integrator'],
                                              out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

                    cpd = prob.check_partials(compact_print=True, method='fd',
                                              includes=['traj0.phases.phase0.integrator'],
                                              out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    def test_brachistochrone_explicit_shooting_path_constraint(self):

        dymos_options['include_check_partials'] = True

        input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=True)

        output_grids = {'same': input_grid,
                        'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7, compressed=True),
                        'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                        'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            with self.subTest(f'output_grid = {output_grid_type}'):
                prob = om.Problem()

                phase = dm.Phase(ode_class=BrachistochroneODE,
                                 transcription=dm.ExplicitShooting(grid=input_grid,
                                                                   output_grid=output_grids[output_grid_type]))

                traj = prob.model.add_subsystem('traj0', dm.Trajectory())
                traj.add_phase('phase0', phase)

                prob.driver = om.ScipyOptimizeDriver()

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=9, units='m/s**2', opt=True, lower=9, upper=9.80665)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)
                phase.add_path_constraint('ydot', lower=-100, upper=0)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('traj0.phase0.t_initial', 0.0)
                prob.set_val('traj0.phase0.t_duration', 2)
                prob.set_val('traj0.phase0.initial_states:x', 0.0)
                prob.set_val('traj0.phase0.initial_states:y', 10.0)
                prob.set_val('traj0.phase0.initial_states:v', 1.0E-6)
                prob.set_val('traj0.phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                prob.run_driver()

                x = prob.get_val('traj0.phase0.timeseries.x')
                y = prob.get_val('traj0.phase0.timeseries.y')
                ydot = prob.get_val('traj0.phase0.timeseries.ydot')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                self.assertTrue(np.all(ydot < 1.0E-6), msg='Not all elements of path constraint satisfied')
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=True, method='cs', out_stream=None,
                                              excludes=['traj0.phases.phase0.integrator'])
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

                    cpd = prob.check_partials(compact_print=True, method='fd', out_stream=None,
                                              includes=['traj0.phases.phase0.integrator'])
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    def test_brachistochrone_explicit_shooting_path_constraint_polynomial_control(self):

        dymos_options['include_check_partials'] = True

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            for compressed in (True, False):
                for path_rename in (True, False):
                    with self.subTest(f'output_grid = {output_grid_type}  compressed = {compressed}  '
                                      f'rename_path_const = {path_rename}'):
                        print(f'output_grid = {output_grid_type}  compressed = {compressed}  ')
                        prob = om.Problem()

                        input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=compressed)

                        output_grids = {'same': input_grid,
                                        'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7,
                                                                          compressed=True),
                                        'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                                        'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

                        phase = dm.Phase(ode_class=BrachistochroneODE,
                                         transcription=dm.ExplicitShooting(grid=input_grid,
                                                                           output_grid=output_grids[output_grid_type]))

                        traj = prob.model.add_subsystem('traj0', dm.Trajectory())
                        traj.add_phase('phase0', phase)

                        prob.driver = om.ScipyOptimizeDriver()

                        phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                        # automatically discover states
                        phase.set_state_options('x', fix_initial=True)
                        phase.set_state_options('y', fix_initial=True)
                        phase.set_state_options('v', fix_initial=True)

                        phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
                        phase.add_control('theta', order=2, val=45.0, units='deg', opt=True,
                                          lower=1.0E-6, upper=179.9, ref=90., control_type='polynomial')

                        phase.add_boundary_constraint('x', loc='final', equals=10.0)
                        phase.add_boundary_constraint('y', loc='final', equals=5.0)
                        phase.add_path_constraint('ydot', constraint_name='foo' if path_rename else None,
                                                  lower=-100, upper=0)

                        phase.add_objective('time', loc='final')

                        prob.setup(force_alloc_complex=True)

                        prob.set_val('traj0.phase0.t_initial', 0.0)
                        prob.set_val('traj0.phase0.t_duration', 2)
                        prob.set_val('traj0.phase0.initial_states:x', 0.0)
                        prob.set_val('traj0.phase0.initial_states:y', 10.0)
                        prob.set_val('traj0.phase0.initial_states:v', 1.0E-6)
                        prob.set_val('traj0.phase0.parameters:g', 9.80665, units='m/s**2')
                        prob.set_val('traj0.phase0.controls:theta',
                                     phase.interp('theta', ys=[0.01, 50]), units='deg')

                        dm.run_problem(prob)

                        x = prob.get_val('traj0.phase0.timeseries.x')
                        y = prob.get_val('traj0.phase0.timeseries.y')
                        t = prob.get_val('traj0.phase0.timeseries.time')

                        assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-5)
                        assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-5)
                        assert_near_equal(t[-1, ...], 1.807379, tolerance=1.0E-5)

                        with np.printoptions(linewidth=1024):
                            cpd = prob.check_partials(compact_print=True, method='cs', out_stream=None,
                                                      excludes=['traj0.phases.phase0.integrator'])
                            assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

                            cpd = prob.check_partials(compact_print=True, method='fd', out_stream=None,
                                                      includes=['traj0.phases.phase0.integrator'])
                            assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    def test_brachistochrone_explicit_shooting_path_constraint_invalid_renamed(self):

        dymos_options['include_check_partials'] = True

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            prob = om.Problem()

            input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=True)

            output_grids = {'same': input_grid,
                            'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7,
                                                              compressed=True),
                            'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                            'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

            phase = dm.Phase(ode_class=BrachistochroneODE,
                             transcription=dm.ExplicitShooting(grid=input_grid,
                                                               output_grid=output_grids[output_grid_type]))

            phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

            # automatically discover states
            phase.set_state_options('x', fix_initial=True)
            phase.set_state_options('y', fix_initial=True)
            phase.set_state_options('v', fix_initial=True)

            phase.add_parameter('g', val=9.80665, units='m/s**2', opt=False)
            phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

            phase.add_boundary_constraint('x', loc='final', equals=10.0)
            phase.add_boundary_constraint('y', loc='final', equals=5.0)
            phase.add_path_constraint('y', constraint_name='foo', lower=5, upper=100)

            prob.model.add_subsystem('phase0', phase)

            phase.add_objective('time', loc='final')

            with warnings.catch_warnings(record=True) as ctx:
                warnings.simplefilter('always')
                prob.setup(check=True)

            self.assertIn("<class 'openmdao.utils.om_warnings.UnusedOptionWarning'>: "
                          "Option 'constraint_name' on path constraint y is only valid for "
                          "ODE outputs. The option is being ignored.", [str(w.message) for w in ctx])

        dymos_options['include_check_partials'] = False

    def test_explicit_shooting_timeseries_ode_output(self):

        dymos_options['include_check_partials'] = True

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            prob = om.Problem()

            prob.driver = om.ScipyOptimizeDriver()

            input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=True)

            output_grids = {'same': input_grid,
                            'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7,
                                                              compressed=True),
                            'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                            'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

            phase = dm.Phase(ode_class=BrachistochroneODE,
                             transcription=dm.ExplicitShooting(grid=input_grid,
                                                               output_grid=output_grids[output_grid_type]))

            phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

            # automatically discover states
            phase.set_state_options('x', fix_initial=True)
            phase.set_state_options('y', fix_initial=True)
            phase.set_state_options('v', fix_initial=True)

            phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
            phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

            phase.add_boundary_constraint('x', loc='final', equals=10.0)
            phase.add_boundary_constraint('y', loc='final', equals=5.0)

            phase.add_timeseries_output('*')

            prob.model.add_subsystem('phase0', phase)

            phase.add_objective('time', loc='final')

            prob.setup(force_alloc_complex=True)

            prob.set_val('phase0.t_initial', 0.0)
            prob.set_val('phase0.t_duration', 2)
            prob.set_val('phase0.initial_states:x', 0.0)
            prob.set_val('phase0.initial_states:y', 10.0)
            prob.set_val('phase0.initial_states:v', 1.0E-6)
            prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
            prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

            dm.run_problem(prob, run_driver=True)

            x = prob.get_val('phase0.timeseries.x')
            y = prob.get_val('phase0.timeseries.y')
            v = prob.get_val('phase0.timeseries.v')
            theta = prob.get_val('phase0.timeseries.theta', units='rad')
            check = prob.get_val('phase0.timeseries.check')
            t = prob.get_val('phase0.timeseries.time')

            assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
            assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)
            assert_near_equal(t[-1, ...], 1.8016, tolerance=1.0E-2)
            assert_near_equal(check, v / np.sin(theta))

            with np.printoptions(linewidth=1024):
                cpd = prob.check_partials(method='cs', out_stream=None, excludes=['phase0.integrator'])
                assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

                cpd = prob.check_partials(method='fd', out_stream=None, includes=['phase0.integrator'])
                assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

            dymos_options['include_check_partials'] = False

    def test_explicit_shooting_unknown_timeseries(self):

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            prob = om.Problem()

            input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=True)

            output_grids = {'same': input_grid,
                            'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7,
                                                              compressed=True),
                            'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                            'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

            phase = dm.Phase(ode_class=BrachistochroneODE,
                             transcription=dm.ExplicitShooting(grid=input_grid,
                                                               output_grid=output_grids[output_grid_type]))

            phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

            # automatically discover states
            phase.set_state_options('x', fix_initial=True)
            phase.set_state_options('y', fix_initial=True)
            phase.set_state_options('v', fix_initial=True)

            phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
            phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

            phase.add_boundary_constraint('x', loc='final', equals=10.0)
            phase.add_boundary_constraint('y', loc='final', equals=5.0)

            phase.add_timeseries_output('*')
            phase.add_timeseries_output('foo')

            prob.model.add_subsystem('phase0', phase)

            phase.add_objective('time', loc='final')

            msg = "Unable to find the source 'foo' in the ODE."

            with self.assertRaises(ValueError) as e:
                prob.setup()

            self.assertIn(msg, str(e.exception))

    def test_brachistochrone_static_gravity_explicit_shooting(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        for output_grid_type in ('same', 'more_dense', 'radau', 'uniform'):
            p = om.Problem()

            input_grid = dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=3, compressed=True)

            output_grids = {'same': input_grid,
                            'more_dense': dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=7,
                                                              compressed=True),
                            'radau': dm.RadauGrid(num_segments=3, nodes_per_seg=4, compressed=True),
                            'uniform': dm.UniformGrid(num_segments=3, nodes_per_seg=11)}

            phase = dm.Phase(ode_class=BrachistochroneODE,
                             ode_init_kwargs={'static_gravity': True},
                             transcription=dm.ExplicitShooting(grid=input_grid,
                                                               output_grid=output_grids[output_grid_type]))

            traj = p.model.add_subsystem('traj', dm.Trajectory())

            traj.add_phase('phase0', phase)

            #
            # Set the variables
            #
            phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

            phase.add_state('x', rate_source='xdot',
                            targets=None,
                            units='m',
                            fix_initial=True)

            phase.add_state('y', rate_source='ydot',
                            targets=None,
                            units='m',
                            fix_initial=True)

            phase.add_state('v', rate_source='vdot',
                            targets=['v'],
                            units='m/s',
                            fix_initial=True)

            phase.add_control('theta', targets=['theta'],
                              continuity=True, rate_continuity=True,
                              units='deg', lower=0.01, upper=179.9)

            phase.add_parameter('g', targets=['g'], static_target=True, opt=False)

            #
            # Constrain the final values of x and y
            #
            phase.add_boundary_constraint('x', loc='final', equals=10)
            phase.add_boundary_constraint('y', loc='final', equals=5)

            #
            # Minimize time at the end of the phase
            #
            phase.add_objective('time', loc='final', scaler=10)

            #
            # Set the optimization driver
            #
            p.driver = om.ScipyOptimizeDriver()

            #
            # Setup the Problem
            #
            p.setup()

            #
            # Set the initial values
            # The initial time is fixed, and we set that fixed value here.
            # The optimizer is allowed to modify t_duration, but an initial guess is provided here.
            #
            p.set_val('traj.phase0.t_initial', 0.0)
            p.set_val('traj.phase0.t_duration', 2.0)

            # Guesses for states are provided at all state_input nodes.
            # We use the phase.interpolate method to linearly interpolate values onto the state input nodes.
            # Since fix_initial=True for all states and fix_final=True for x and y, the initial or final
            # values of the interpolation provided here will not be changed by the optimizer.
            p.set_val('traj.phase0.initial_states:x', phase.interp('x', [0, 10]))
            p.set_val('traj.phase0.initial_states:y', phase.interp('y', [10, 5]))
            p.set_val('traj.phase0.initial_states:v', phase.interp('v', [0, 9.9]))

            # Guesses for controls are provided at all control_input node.
            # Here phase.interpolate is used to linearly interpolate values onto the control input nodes.
            p.set_val('traj.phase0.controls:theta', phase.interp('theta', [5, 100.5]))

            # Set the value for gravitational acceleration.
            p.set_val('traj.phase0.parameters:g', 9.80665)

            #
            # Solve for the optimal trajectory
            #
            dm.run_problem(p, simulate=False)

            assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)



import numpy as np

import openmdao.api as om


class ThrustComp(om.ExplicitComponent):
    """ Computes mass flow rate for the F4's 2 J79 engines at full throttle. """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('max_thrust', shape=(nn,), desc='maximum thrust', units='N')
        self.add_input('throttle', val=np.ones(nn), desc='throttle parameter', units=None)

        # Outputs
        self.add_output(name='thrust', val=np.zeros(nn),
                        desc='vehicle thrust at given throttle value',
                        units='N')

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='thrust', wrt='max_thrust', rows=ar, cols=ar)
        self.declare_partials(of='thrust', wrt='throttle', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['thrust'] = inputs['max_thrust'] * inputs['throttle']

    def compute_partials(self, inputs, partials):
        partials['thrust', 'max_thrust'] = inputs['throttle']
        partials['thrust', 'throttle'] = inputs['max_thrust']



import openmdao.api as om


class FlyingRobotODE(om.ExplicitComponent):
    """
    The flying robot ODE mimics a free flying robot with thrusters in the vertical axis and
    horizontal axis.

    The controls on the robot are external accelerations in these two orthogonal axes.
    Gravity acts in the negative axis.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('u', shape=(nn, 2), units='m/s**2')
        self.add_output('u_mag2', shape=(nn,), units='m**2/s**4', desc='square of control magnitude')

        self.declare_coloring(wrt='u', method='cs')
        self.declare_partials(of='u_mag2', wrt='u', method='cs')

    def compute(self, inputs, outputs):
        u_x = inputs['u'][:, 0]
        u_y = inputs['u'][:, 1]

        outputs['u_mag2'] = u_x**2 + u_y**2
